/*
  This file is part of TALER
  Copyright (C) 2023-2026 Taler Systems SA

  TALER is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2.1,
  or (at your option) any later version.

  TALER is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General
  Public License along with TALER; see the file COPYING.LGPL.
  If not, see <http://www.gnu.org/licenses/>
*/
/**
 * @file src/lib/merchant_api_post-private-accounts.c
 * @brief Implementation of the POST /private/accounts request
 * @author Christian Grothoff
 */
#include "platform.h"
#include <curl/curl.h>
#include <jansson.h>
#include <microhttpd.h> /* just for HTTP status codes */
#include <gnunet/gnunet_util_lib.h>
#include <gnunet/gnunet_curl_lib.h>
#include <taler/merchant/post-private-accounts.h>
#include "merchant_api_curl_defaults.h"
#include "merchant_api_common.h"
#include <taler/taler_json_lib.h>
#include <taler/taler_curl_lib.h>


/**
 * Handle for a POST /private/accounts operation.
 */
struct TALER_MERCHANT_PostPrivateAccountsHandle
{
  /**
   * Base URL of the merchant backend.
   */
  char *base_url;

  /**
   * The full URL for this request.
   */
  char *url;

  /**
   * Handle for the request.
   */
  struct GNUNET_CURL_Job *job;

  /**
   * Function to call with the result.
   */
  TALER_MERCHANT_PostPrivateAccountsCallback cb;

  /**
   * Closure for @a cb.
   */
  TALER_MERCHANT_POST_PRIVATE_ACCOUNTS_RESULT_CLOSURE *cb_cls;

  /**
   * Reference to the execution context.
   */
  struct GNUNET_CURL_Context *ctx;

  /**
   * Minor context that holds body and headers.
   */
  struct TALER_CURL_PostContext post_ctx;

  /**
   * Payto URI to add.
   */
  struct TALER_FullPayto payto_uri;

  /**
   * Optional credit facade URL.
   */
  char *credit_facade_url;

  /**
   * Optional credit facade credentials (JSON).
   */
  json_t *credit_facade_credentials;

  /**
   * Optional extra wire subject metadata (JSON).
   */
  char *extra_wire_subject_metadata;
};


/**
 * Function called when we're done processing the
 * HTTP POST /private/accounts request.
 *
 * @param cls the `struct TALER_MERCHANT_PostPrivateAccountsHandle`
 * @param response_code HTTP response code, 0 on error
 * @param response response body, NULL if not in JSON
 */
static void
handle_post_account_finished (void *cls,
                              long response_code,
                              const void *response)
{
  struct TALER_MERCHANT_PostPrivateAccountsHandle *pah = cls;
  const json_t *json = response;
  struct TALER_MERCHANT_PostPrivateAccountsResponse apr = {
    .hr.http_status = (unsigned int) response_code,
    .hr.reply = json
  };

  pah->job = NULL;
  GNUNET_log (GNUNET_ERROR_TYPE_INFO,
              "POST /private/accounts completed with response code %u\n",
              (unsigned int) response_code);
  switch (response_code)
  {
  case 0:
    apr.hr.ec = TALER_EC_GENERIC_INVALID_RESPONSE;
    break;
  case MHD_HTTP_OK:
    {
      struct GNUNET_JSON_Specification spec[] = {
        GNUNET_JSON_spec_fixed_auto ("h_wire",
                                     &apr.details.ok.h_wire),
        GNUNET_JSON_spec_fixed_auto ("salt",
                                     &apr.details.ok.salt),
        GNUNET_JSON_spec_end ()
      };

      if (GNUNET_OK !=
          GNUNET_JSON_parse (json,
                             spec,
                             NULL, NULL))
      {
        GNUNET_break_op (0);
        apr.hr.http_status = 0;
        apr.hr.ec = TALER_EC_GENERIC_INVALID_RESPONSE;
        break;
      }
    }
    break;
  case MHD_HTTP_ACCEPTED:
    if (GNUNET_OK !=
        TALER_MERCHANT_parse_mfa_challenge_response_ (
          json,
          &apr.details.accepted))
    {
      GNUNET_break_op (0);
      apr.hr.http_status = 0;
      apr.hr.ec = TALER_EC_GENERIC_INVALID_RESPONSE;
    }
    break;
  case MHD_HTTP_BAD_REQUEST:
    GNUNET_break_op (0);
    apr.hr.ec = TALER_JSON_get_error_code (json);
    apr.hr.hint = TALER_JSON_get_error_hint (json);
    break;
  case MHD_HTTP_FORBIDDEN:
    apr.hr.ec = TALER_JSON_get_error_code (json);
    apr.hr.hint = TALER_JSON_get_error_hint (json);
    break;
  case MHD_HTTP_NOT_FOUND:
    apr.hr.ec = TALER_JSON_get_error_code (json);
    apr.hr.hint = TALER_JSON_get_error_hint (json);
    break;
  case MHD_HTTP_CONFLICT:
    apr.hr.ec = TALER_JSON_get_error_code (json);
    apr.hr.hint = TALER_JSON_get_error_hint (json);
    break;
  case MHD_HTTP_INTERNAL_SERVER_ERROR:
    apr.hr.ec = TALER_JSON_get_error_code (json);
    apr.hr.hint = TALER_JSON_get_error_hint (json);
    break;
  default:
    TALER_MERCHANT_parse_error_details_ (json,
                                         response_code,
                                         &apr.hr);
    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
                "Unexpected response code %u/%d\n",
                (unsigned int) response_code,
                (int) apr.hr.ec);
    GNUNET_break_op (0);
    break;
  }
  pah->cb (pah->cb_cls,
           &apr);
  if (MHD_HTTP_ACCEPTED == response_code)
    TALER_MERCHANT_mfa_challenge_response_free (
      &apr.details.accepted);
  TALER_MERCHANT_post_private_accounts_cancel (pah);
}


struct TALER_MERCHANT_PostPrivateAccountsHandle *
TALER_MERCHANT_post_private_accounts_create (
  struct GNUNET_CURL_Context *ctx,
  const char *url,
  struct TALER_FullPayto payto_uri)
{
  struct TALER_MERCHANT_PostPrivateAccountsHandle *pah;

  pah = GNUNET_new (struct TALER_MERCHANT_PostPrivateAccountsHandle);
  pah->ctx = ctx;
  pah->base_url = GNUNET_strdup (url);
  pah->payto_uri.full_payto = GNUNET_strdup (payto_uri.full_payto);
  return pah;
}


enum GNUNET_GenericReturnValue
TALER_MERCHANT_post_private_accounts_set_options_ (
  struct TALER_MERCHANT_PostPrivateAccountsHandle *pah,
  unsigned int num_options,
  const struct TALER_MERCHANT_PostPrivateAccountsOptionValue *options)
{
  for (unsigned int i = 0; i < num_options; i++)
  {
    switch (options[i].option)
    {
    case TALER_MERCHANT_POST_PRIVATE_ACCOUNTS_OPTION_END:
      return GNUNET_OK;
    case TALER_MERCHANT_POST_PRIVATE_ACCOUNTS_OPTION_CREDIT_FACADE_URL:
      GNUNET_free (pah->credit_facade_url);
      pah->credit_facade_url = GNUNET_strdup (options[i].details.credit_facade_url);
      break;
    case TALER_MERCHANT_POST_PRIVATE_ACCOUNTS_OPTION_CREDIT_FACADE_CREDENTIALS:
      json_decref (pah->credit_facade_credentials);
      pah->credit_facade_credentials
        = json_incref ((json_t *) options[i].details.credit_facade_credentials);
      break;
    case TALER_MERCHANT_POST_PRIVATE_ACCOUNTS_OPTION_EXTRA_WIRE_SUBJECT_METADATA:
      GNUNET_free (pah->extra_wire_subject_metadata);
      pah->extra_wire_subject_metadata
        = GNUNET_strdup (options[i].details.extra_wire_subject_metadata);
      break;
    default:
      GNUNET_break (0);
      return GNUNET_SYSERR;
    }
  }
  return GNUNET_OK;
}


enum TALER_ErrorCode
TALER_MERCHANT_post_private_accounts_start (
  struct TALER_MERCHANT_PostPrivateAccountsHandle *pah,
  TALER_MERCHANT_PostPrivateAccountsCallback cb,
  TALER_MERCHANT_POST_PRIVATE_ACCOUNTS_RESULT_CLOSURE *cb_cls)
{
  json_t *req_obj;
  CURL *eh;

  pah->cb = cb;
  pah->cb_cls = cb_cls;
  pah->url = TALER_url_join (pah->base_url,
                             "private/accounts",
                             NULL);
  if (NULL == pah->url)
    return TALER_EC_GENERIC_CONFIGURATION_INVALID;
  req_obj = GNUNET_JSON_PACK (
    TALER_JSON_pack_full_payto (
      "payto_uri",
      pah->payto_uri),
    GNUNET_JSON_pack_allow_null (
      GNUNET_JSON_pack_string (
        "credit_facade_url",
        pah->credit_facade_url)),
    GNUNET_JSON_pack_allow_null (
      GNUNET_JSON_pack_object_incref (
        "credit_facade_credentials",
        pah->credit_facade_credentials)),
    GNUNET_JSON_pack_allow_null (
      GNUNET_JSON_pack_string (
        "extra_wire_subject_metadata",
        pah->extra_wire_subject_metadata)));
  eh = TALER_MERCHANT_curl_easy_get_ (pah->url);
  if ( (NULL == eh) ||
       (GNUNET_OK !=
        TALER_curl_easy_post (&pah->post_ctx,
                              eh,
                              req_obj)) )
  {
    GNUNET_break (0);
    json_decref (req_obj);
    if (NULL != eh)
      curl_easy_cleanup (eh);
    return TALER_EC_GENERIC_INTERNAL_INVARIANT_FAILURE;
  }
  json_decref (req_obj);
  pah->job = GNUNET_CURL_job_add2 (pah->ctx,
                                   eh,
                                   pah->post_ctx.headers,
                                   &handle_post_account_finished,
                                   pah);
  if (NULL == pah->job)
    return TALER_EC_GENERIC_INTERNAL_INVARIANT_FAILURE;
  return TALER_EC_NONE;
}


void
TALER_MERCHANT_post_private_accounts_cancel (
  struct TALER_MERCHANT_PostPrivateAccountsHandle *pah)
{
  if (NULL != pah->job)
  {
    GNUNET_CURL_job_cancel (pah->job);
    pah->job = NULL;
  }
  TALER_curl_easy_post_finished (&pah->post_ctx);
  GNUNET_free (pah->payto_uri.full_payto);
  GNUNET_free (pah->url);
  GNUNET_free (pah->credit_facade_url);
  json_decref (pah->credit_facade_credentials);
  GNUNET_free (pah->extra_wire_subject_metadata);
  GNUNET_free (pah->base_url);
  GNUNET_free (pah);
}


/* end of merchant_api_post-private-accounts.c */
